

#local


#code PAGE0

; pseudo random number generator seed
; must not be zero
rndseed16:: dw	$4711		


#code VCC_INT
#include "vars16.s"



; ______________________________________
; RSWAPWITHVAR ( T T* -- T )

 macro RSWAPWITHVAR
	pop	hl		; meine var
	call	rswapwithvar
 endm	

rswapwithvar::
	ex	hl,de		; hl -> var	
nswapwithvar::
	ld	a,(hl)
	ld	(hl++),e
	ld	e,a
	
	ld	a,(hl)
	ld	(hl),d
	ld	d,a
	
	ret

; ______________________________________
; NSWAPWITHVAR ( T* T -- T )

 macro NSWAPWITHVAR
	pop	hl		; hl -> var
	call	nswapwithvar
 endm



; -----------------------------------------
; Arrays and Structs:
; -----------------------------------------
;
; variables are pointers to handles
; handles are pointers to payload data
; struct / array length.w is at payload-2




; -----------------------------------------
; Arithmetics & Logics:
; -----------------------------------------

; ________________________________
; increment value
;
; INCR ( n -- n )

 macro INCR
 	inc	de
 endm

; ________________________________
; decrement value
;
; DECR ( n -- n )

 macro DECR
 	dec	de
 endm

; ________________________________
; calculate absolute value of signed int
; return value is uint16, because abs($8000) = $8000
;
; ABS ( n -- n )

 macro ABS
	call	abs
 endm

; ________________________________
; negate
;
; NEG ( n -- n )

 macro NEG16		; NEG is a Z80 instr name
 	call	neg
 endm

abs::	bit	7,d
	ret	z
neg:: 	xor	a
	sub	a,e
	ld	e,a
	sbc	a,a	; nc => 0, cy => 0xff
	sub	a,d
	ld	d,a
	ret

; ________________________________
; complement
;
; CPL ( n -- n )

 macro CPL16		; CPL is a Z80 instr name
 	call	cpl
 endm
 
cpl32::	ld	a,h
	cpl
	ld	h,a
	ld	a,l
	cpl	
	ld	l,a

cpl:: 	ld	a,d
	cpl
	ld	d,a
	ld	a,e
	cpl
	ld	e,a
	ret

; ________________________________
; boolean negation: !n
;
; NOT ( n -- n )

 macro NOT
 	call	not
 endm
 
not:: 	ld	a,e
	inc	e	; assume de=0 -> de=1
	or	d
	ret	z	

return_0::
	ld	de,0
	ret

; ________________________________
; calculate sign of signed int
;
; SIGN ( n -- n )

 macro SIGN
 	call	sign
 endm
 
sign::	ld	a,d
	add	a
	jr	c,return_minus1
	or	e
	ret	z	; 0

return_1::
	ld	de,1	; +1
	ret

return_minus1::
	ld	de,-1
	ret

; ________________________________
; convert to boolean
;
; BOOL ( n -- n )

 macro BOOL
 	call	bool
 endm
 
bool::	ld	a,d
	or	e
	ret	z
	ld	de,1
	ret

; ________________________________
; shift left once
;
; SL1 ( n -- n )

 macro SL1
 	ex	de,hl
	add	hl,hl
	ex	de,hl
 endm

; ________________________________
; signed shift right once
;
; SR1s ( n -- n )

 macro SR1s
 	sra	d
	rr	e
 endm

; ________________________________
; unsigned shift right once
;
; SR1u ( n -- n )

 macro SR1u
 	srl	d
	rr	e
 endm

; ________________________________
; unsigned shift right two digits
;
; SR2u ( n -- n )

 macro SR2u
 	call	sr2u
 endm
 
sr2u:: 	srl	d
	rr	e
	srl	d
	rr	e
	ret
	
; ________________________________
; add
;
; ADD    ( n n -- n )
; ADDp8i ( b* n -- b* )		; add offset to byte pointer

 macro ADD16
	pop	hl
	add	hl,de
	ex	hl,de
 endm

; ________________________________
; subtract n1 - n2
;
; SUB    ( n1 n2 -- n )
; SUBp8p ( b* b* -- n )		; subtract two byte pointers
; SUBp8i ( b* n -- b* )		; subtract offset from byte pointer

 macro SUB16
 	pop	hl		; hl=n1, de=n2
	and	a
	sbc	hl,de
	ex	hl,de
  endm

; ________________________________
; add offset to word pointer
;
; ADDp16i ( w* n -- w* )

 macro ADDp16i
	pop	hl	
	ex	hl,de
	add	hl,hl		; hl = offset*2
	add	hl,de
	ex	de,hl
 endm

; ________________________________
; subtract offset from word pointer
;
; SUBp16i ( w* n -- w* )	

 macro SUBp16i
	pop	hl
	call	subp16i
 endm	

subp16i::
	ex	hl,de
	add	hl,hl		
	ex	hl,de		; de = offset*2
	and	a
	sbc	hl,de
	ex	de,hl
	ret

; ________________________________
; subtract two word pointers
; the result is signed
;
; SUBp16p ( w1* w2* -- i16 )	

 macro SUBp16p
	pop	hl
	call	subp16p
 endm
	
subp16p::
	and	a
	sbc	hl,de		; hl = p1-p2
	ex	hl,de		; de = p1-p2
	sra	d
	rr	e
	ret

; ________________________________
; add immediate value
; also used for SUBI
;
; ADDI ( n $value.w -- n )

 macro ADDI $value
 	ld	hl,$value
	add	hl,de
	ex	hl,de
 endm	

; ________________________________
; bitwise OR
;
; OR ( n n -- n )

 macro OR16
 	pop	hl
	call	or
 endm
 	
or::	ld	a,h
	or	d
	ld	d,a
	ld	a,l
	or	e
	ld	e,a
	ret

; ________________________________
; bitwise XOR
;
; XOR ( n n -- n )

 macro XOR16
 	pop	hl
	call	xor
 endm
 
xor::	ld	a,h
	xor	d
	ld	d,a
	ld	a,l
	xor	e
	ld	e,a
	ret

; ________________________________
; bitwise AND
;
; AND ( n n -- n )

 macro AND16
 	pop	hl
	call	and
 endm
 	
and::	ld	a,h
	and	d
	ld	d,a
	ld	a,l
	and	e
	ld	e,a
	ret

; ________________________________
; get high-byte of word
;
; GETHI ( n -- n )

 macro GETHI
 	ld	e,d
	ld	d,0
 endm	

; ________________________________
; get low-byte of word
;
; GETLO ( n -- n )

 macro GETLO
 	ld	d,0
 endm

; ________________________________
; swap high and low byte
;
; SWAPBYTES ( n -- n )

 macro SWAPBYTES
	ld	a,d
	ld	d,e
	ld	e,a
 endm

; ________________________________
; calculate unsigned maximum
;
; MAXu ( n n -- n )

 macro MAXu
 	pop	hl
	call	maxu
 endm
 	
maxu::	and	a
	sbc	hl,de		
	ret	c		; hl-de<0 => hl<de => max=de => fertig
	add	hl,de		; else: max=hl => hl restaurieren
	ex	de,hl		; 	  und nach de laden
	ret

; ________________________________
; calculate unsigned minimum
;
; MINu ( n n -- n )

 macro MINu
 	pop	hl
	call	minu
 endm
 
minu::	and	a
	sbc	hl,de
	ret	nc		; hl-de≥0 => hl≥de => min=de => fertig
	add	hl,de		; else: min=hl => hl restaurieren
	ex	de,hl		; 	  und nach de laden
	ret

; ________________________________
; calculate the most significant bit:
; the bit number of the highest set '1' bit
; returns 15 .. 0
; returns 0 for $0001 and for $0000
;
; MSBIT ( n -- n )

 macro MSBIT
 	call	msbit
 endm
 
msbit::	ld	a,d
	and	a
	jr	nz,1$		; falls ein Bit in uint16.hi gesetzt

	ld	a,e
	and	a
	ret	z		; msbit(0) = 0

	ld	de,7		; msbit = max. 7 wenn ein Bit in uint16.lo gesetzt
	jr	2$

1$:	ld	de,16		; msbit = max. 15 wenn ein Bit in uint16.hi gesetzt

3$:	dec	e
2$:	rla
	jr	nc,3$
	ret

; ________________________________
; shift left n1 << ival
; shift must be in range 2 to 15
;
; SLN ( n %shift -- n )

 macro SLN %shift
 	ex	hl,de
	call	sl%shift
	ex	hl,de
 endm

sl15::	add	hl,hl
sl14::	add	hl,hl
sl13::	add	hl,hl
sl12::	add	hl,hl
sl11::	add	hl,hl
sl10::	add	hl,hl
sl9::	add	hl,hl
sl8::	ld	h,l
	ld	l,0
	ret
 
sl7::	add	hl,hl
sl6::	add	hl,hl
sl5::	add	hl,hl
sl4::	add	hl,hl
sl3::	add	hl,hl
sl2::	add	hl,hl
sl1::	add	hl,hl
sl0::	ret


; ________________________________
; shift left n1 << n2
;
; SL ( n1 n2 -- n )

 macro SL
 	pop	hl		; hl=n1, de=shift
	call	slx
 endm	
	
; ________________________________
; slx: de = hl << e
; sl:  de = de << a

slx::	ex	hl,de		; de=n1, hl=shift
	ld	a,l
sl:	and	a,$F
	ret	z

	cp	a,8
	jr	c,3$

	ld	d,e
	ld	e,0
	ret	z
	sub	a,8

3$:	ex	hl,de
2$:	add	hl,hl
	dec	a
	jr	nz,2$
	ex	hl,de
	ret

; ________________________________
; unsigned shift right n1 >> n2
;
; SRu ( n1 n2 -- n )

 macro SRu
 	pop	hl
	call	srux
 endm
	
; ________________________________
; srux: de = hl >> e
; sru:  de = de >> a

srux::	ex	hl,de		; de=n1, hl=shift
	ld	a,l
sru:	and	a,$F
	ret	z

	cp	a,8
	jr	c,sr2

sr1:	ld	e,d
	ld	d,0
	ret	z
	sub	a,8

sr2:	srl	d
	rr	e
	dec	a
	jr	nz,sr2
	ret

; ________________________________
; signed shift right n1 >> n2
;
; SRs ( n1 n2 -- n )

 macro SRs
 	pop	hl
	call	srsx
 endm
 
; ________________________________
; srsx: de = hl >> e
; srs:  de = de >> l

srsx::	ex	hl,de		; de=n1, hl=shift
	ld	a,l
srs:	and	a,$F
	ret	z

	cp	a,8
	jr	c,2$

	bit	7,d
	jr	z,sr1

	ld	e,d
	ld	d,0xff

	sub	a,8
	ret	z

2$:	sra	d
	rr	e
	dec	a
	jr	nz,2$
	ret

; ________________________________
; multiply n1*n2
;
; MUL ( n1 n2 -- n )

 macro MUL
 	pop	hl
	call	mul_hl_de
	ex	hl,de
 endm

; ________________________________
; multiply signed / unsigned
;
; HL = HL * DE		<-- Result is in HL!
;
; pres: BC, DE

mul_hl_de::
	xor 	a
	cp	a,h
	jr	z,lxde	; h=0 => l*de
	cp	a,d
	jr	z,hlxe	; d=0 => hl*e

; HL≥256 && DE≥256:	; => lot of overflow

	ld	a,l
	push	af	; l
	ld	a,h
	call	axel	; hl = h*de, de=pres
	pop	af	; a=l
	jr	axel	; hl = (h*de)<<8 + l*de


hlxe: ex	de,hl	; HL = HL * E
lxde: ld	a,l	; HL = DE * L
axde: ld	l,0	; HL = A * DE
axel::			; HL = L<<8 + A * DE

	scf		; -> set bit 0
	adc	a	; -> cy, z
	jr	c,3$

1$:	add	hl,hl
2$:	add	a	; -> cy, z
	jr	nc,1$
	ret	z	; this was the stopper
3$:	add	hl,hl
	add	hl,de
	jr	2$


; ________________________________
; divide unsigned n1/n2
;
; DIVu ( n1 n2 -- n )

 macro DIVu
 	pop	hl
	call	divu
 endm	

; ________________________________
; remainder of unsigned n1/n2
;
; REMu ( n1 n2 -- n )

 macro REMu
 	pop	hl
	call	divu
	ex	hl,de
 endm

; ________________________________
; divide signed n1/n2
;
; DIVs ( n1 n2 -- n )

 macro DIVs
 	pop	hl
	call	divs
 endm

; ________________________________
; remainder of signed n1/n2
;
; REMs ( n1 n2 -- n )

 macro REMs
 	pop	hl
	call	rems
	ex	hl,de
 endm

; ________________________________
; remainder of signed division
; the remainder has the same sign as HL
;
; HL = HL % DE

rems::	bit	7,h
	jr	z,div0		; HL pos => result pos
	call	div1		; HL neg => divide and negate result
neg_hl::
 	xor	a
	sub	a,l
	ld	l,a
	sbc	a,a		; nc => 0, cy => 0xff
	sub	a,h
	ld	h,a
	ret

; ________________________________
; signed division
;
; DE = HL / DE

divs::	ld	a,h
	or	d
	jp	p,divu		; both pos.

	ld	a,h
	xor	d
	jp	p,div		; both neg. => result pos.
	call	div		; else divide & neg result
	jp	neg

div: 	bit	7,h
div1: 	call	nz,neg_hl
div0: 	bit	7,d
	call	nz,neg
	;jr	divu

; ________________________________
; unsigned 16-bit division
;
; DE = HL / DE
; HL = HL % DE
;
; If divisor is 0, quotient is set to "infinity", i.e HL = 0xFFFF.

; Two algorithms:
; one assumes divisor < 2^7,
; the second assumes divisor ≥ 2^7;
; Both algorithms "rotate" 24 bits (H,L,A) but roles change.
; choose the applicable one:

divu::	ld	a,e
	and	a,$80
	or	a,d
	jr	nz,morethan7bits

; ________________________________
; unsigned division
; divisor < 2^7
;
; divu7:	DE = HL / DE  &  HL = HL % DE	 	unsigned	if de ≤ 0x80

divu7::
	ld	d,16		; bits in dividend and possible quotient
	add	hl,hl

dvloop7:
	; HL holds both dividend and quotient. While we shift a bit from
	; MSB of dividend, we shift next bit of quotient in from carry.
	; A holds remainder.
	rla

	; If remainder is ≥ divisor, next bit of quotient is 1:
	sub	a,e		; try to subtract
	jr	nc,nodrop7	; success
	add	a,e       	; else undo, sets cy because sub a,e did set it

nodrop7:
	ccf			; cy = next bit of quotient
	adc	hl,hl		; cy = next bit from current remainder of divident
	dec	d		; preserves cy
	jr	nz,dvloop7

	; Carry now contains the same value it contained before
	; entering dvloop7[*]: "0" = valid result.
	ld	e,a		; DE = remainder, HL = quotient
	ex	de,hl		; DE = quotient,  HL = remainder
	ret

; ________________________________
; unsigned division
; divisor ≥ 2^7

morethan7bits:
	push	bc
	ld	b,9		; at most 9 bits in quotient.
	ld	a,l		; precompute the first 7 shifts
	ld	l,h		; by doing 8
	ld	h,0
	rr	l		; and undoing 1

dvloop:
	; Shift next bit of quotient into bit 0 of dividend
	; Shift next MSB of dividend into LSB of remainder
	; A holds both dividend and quotient. While we shift a bit from
	; MSB of dividend, we shift next bit of quotient in from carry
	; HL holds remainder
	adc	hl,hl		; HL < 2^(7+9), no carry, ever.

	; If remainder is >= divisor, next bit of quotient is 1.
	sbc	hl,de		; try to subtract
	jr	nc,nodrop	; success
	add	hl,de		; else undo
				; this also sets the carry, because sbc hl,de did set it.
nodrop:
	ccf			; cy = next bit of quotient
	rla
	djnz    dvloop

	; Take care of the ninth quotient bit! after the loop B=0.
	rl	b		; BA = quotient; also clears cy
	ld	d,b
	ld	e,a		; DE = quotient, HL = remainder
	pop	bc		; restore
	ret			; nc




; -----------------------------------------
; 	Unsigned Compare
; -----------------------------------------

; ______________________________________
; unsigned compare: n1>n2
;
; GTu ( n1 n2 -- n )

 macro GTu
 	pop	hl
	call	gtu
 endm
 

; ______________________________________
; unsigned compare: n1<n2
;
; LTu ( n1 n2 -- n )

 macro LTu
 	pop	hl		; n1
	call	ltu
 endm
 
gtu::	ex	de,hl		; hl=n2, de=n1	
ltu::	xor	a		; a=0, cy=0
	sbc	hl,de		; cy => n1-n2<0 => n1<n2 => lt=yes

	ld	d,a		; d=0
	adc	a,a		; nc=>0, cy=>1
	ld	e,a		; de=0|1
	ret

; ______________________________________
; unsigned compare: n1≤n2
;
; LEu ( n1 n2 -- n )

 macro LEu
 	pop	hl
	call	leu
 endm
 

; ______________________________________
; unsigned compare: n1≥n2
;
; GEu ( n1 n2 -- n )

 macro GEu
 	pop	hl
	call	geu
 endm
 	
leu::	ex	de,hl		; hl=n2, de=n1	
geu::	xor	a		; a=0, cy=0
	sbc	hl,de		; n1-n2

ld_de_nc::			; in: a=0 !!!
	ccf			; cy => n1-n2<0 => n1<n2 => ge=no => ccf
	
ld_de_cy::			; in: a=0 !!!
	ld	d,a		; d=0
	adc	a,a		; nc=>0, cy=>1
	ld	e,a		; de=0|1
	ret

; ______________________________________
; compare two values: n1 == n2
; either signed - signed or
; unsigned - unsigned
;
; EQ ( n n -- n )

 macro EQ	
 	pop	hl
	call	eq
 endm
 	
eq::	xor	a
	sbc	hl,de
	ld	d,a
	ld	e,a		; de=0
	ret	nz		; eq=false
	inc	e
	ret			; eq=true

; ______________________________________
; compare two values: n1 != n2
; either signed - signed or
; unsigned - unsigned
;
; NE ( n n -- n )

 macro NE
 	pop	hl
	call	ne
 endm
 
ne::	xor	a
	sbc	hl,de
	ld	d,a
	ld	e,a
	ret	z		; ne=false
	inc	e
	ret			; ne=true


; -----------------------------------------
; 	Signed Compare
; -----------------------------------------

; _______________________________________
; signed compare: n1>n2
; signed compare: n1<n2
;
; GTs ( n1 n2 -- n )
; LTs ( n1 n2 -- n )

 macro GTs
 	pop	hl		; hl=n1, de=n2, op: hl>de
	call	gts
 endm
 	
 macro LTs
 	pop	hl		; hl=n1, de=n2, op: hl<de
	call	lts
 endm

gts::	ex	hl,de		; op: hl<de
lts::	ld	a,h
	xor	d
	jp	p,ltu		; both neg / both pos => use LTu

	xor	a
	add	hl,hl		; cy = hl.vz
	jr	ld_de_cy	; result true if hl<0

; _______________________________________
; signed compare: n1≤n2
; signed compare: n1≥n2
;
; LEs ( n1 n2 -- n )
; GEs ( n1 n2 -- n )

 macro LEs
 	pop	hl
	call	les
 endm

 macro GEs
 	pop	hl
	call	ges
 endm	
 	
les::	ex	de,hl		; op: hl≥de
ges::	ld	a,h		
	xor	d
	jp	p,geu		; both neg / both pos => use GEu

	xor	a
	add	hl,hl		; cy = hl.vz
	jr	ld_de_nc	; result false if hl<0


; -----------------------------------------
; 	Mixed Signedness Compare
; -----------------------------------------

; ______________________________________
; compare two values with mixed signedness
; either signed - unsigned or vice versa
; 
; EQsu ( n n -- f )
; EQus ( n n -- f )

 macro EQx
 	pop	hl
	call	eqx
 endm
 	
eqx::	ld	a,d
	or	h		; either of both exceeds common range?
	jp	p,eq		; no
return_false:			; yes => can't be equal
	ld	de,0
	ret

; ______________________________________
; compare two values with mixed signedness
; either signed - unsigned or vice versa
; 
; NEsu ( n n -- f )
; NEus ( n n -- f )

 macro NEx
 	pop	hl
	call	nex
 endm
 	
nex::	ld	a,d
	or	h		; either of both exceeds common range?
	jp	p,ne		; no
return_true:			; yes => must be not equal
	ld	de,1
	ret

; ______________________________________
; compare two values with mixed signedness u16 > i16
; compare two values with mixed signedness i16 < u16
;
; GTus ( u16 i16 -- f )
; LTsu ( i16 u16 -- f )

 macro GTus
 	pop	hl
	call	gtus
 endm

 macro LTsu
 	pop	hl
	call	ltsu
 endm
 	
gtus::	ex	hl,de
ltsu:	ld	a,h		; i16 < 0 ?
	or	d		; u16 ≥ 0x8000 ?
	jp	p,ltu		; no
	jr	return_true	; yes => i16 is less than u16
	
; ______________________________________
; compare two values with mixed signedness i16 > u16
; compare two values with mixed signedness u16 < i16
;
; GTsu ( i16 u16 -- f )
; LTus ( u16 i16 -- f )

 macro GTsu
 	pop	hl
	call	gtsu
 endm

 macro LTus
 	pop	hl
	call	ltus
 endm	
 	
gtsu::	ex	hl,de
ltus:	ld	a,h		; u16 ≥ 0x8000 ?
	or	d		; i16 < 0 ?
	jp	p,ltu
	jr	return_false	; yes => u16 can't be less than i16

; ______________________________________
; compare two values with mixed signedness u16 ≤ i16
; compare two values with mixed signedness i16 ≥ u16
;
; LEus ( u16 i16 -- f )
; GEsu ( i16 u16 -- f )

 macro LEus
 	pop	hl
	call	leus
 endm
 
 macro GEsu
 	pop	hl
	call	gesu
 endm	
 	
leus::	ex	hl,de
gesu:	ld	a,h		; i16 < 0 ?
	or	d		; u16 ≥ 0x8000 ?
	jp	p,geu		; no
	jr	return_false	; yes => i16 can't be ≥ u16

; ______________________________________
; compare two values with mixed signedness i16 ≤ u16
; compare two values with mixed signedness u16 ≥ i16
;
; LEsu ( i16 u16 -- f )
; GEus ( u16 i16 -- f )

 macro LEsu
 	pop	hl
	call	lesu
 endm

 macro GEus
 	pop	hl	
	call	geus
 endm	
 
lesu::	ex	hl,de
geus:	ld	a,h		; u16 ≥ 0x8000 ?
	or	d		; i16 < 0 ?
	jp	p,geu		; no
	jr	return_true	; yes => u16 must be ≥ i16




; -----------------------------------------
; Assignments
; -----------------------------------------


; _________________________________
; reset bits in variable
;
; ANDGL ( n &var -- )

 macro ANDGL
 	pop	hl
	call	andgl
	pop	de
 endm
 	
andgl::	ld	a,(de)		; de=addr, hl=value
	and	l
	ld	(de++),a
	ld	a,(de)
	and	h
	ld	(de),a
	ret

; _________________________________
; set bits in variable
;
; ORGL ( n &var -- )

 macro ORGL
 	pop	hl
	call	orgl
	pop	de
 endm
 	
orgl::	ld	a,(de)		; de=addr, hl=value
	or	l
	ld	(de++),a
	ld	a,(de)
	or	h
	ld	(de),a
	ret

; _________________________________
; toggle bits in variable
;
; XORGL ( n &var -- )

 macro XORGL
 	pop	hl
	call	xorgl
	pop	de
 endm
 	
xorgl::	ld	a,(de)		; de=addr, hl=value
	xor	l
	ld	(de++),a
	ld	a,(de)
	xor	h
	ld	(de),a
	ret

; _________________________________
; add value to variable
;
; ADDGL ( n &w -- )
; ADDGLp8 ( n &b* -- )
; ADDGLp16 ( n &w* -- )
; ADDGLp32 ( n &l* -- )

 macro ADDGL
 	pop	hl
	call	addgl
	pop	de
 endm
 	
 macro ADDGLp16
 	pop	hl
	call	addglp16
	pop	de
 endm
 	
 macro ADDGLp32
 	pop	hl
	call	addglp32
	pop	de
 endm
 	
addglp32::
	add	hl,hl
addglp16::
	add	hl,hl
addgl::	ld	a,(de)
	add	l
	ld	(de++),a
	ld	a,(de)
	adc	h
	ld	(de),a
	ret

; _________________________________
; subtract value from variable
;
; SUBGL ( n &w -- )
; SUBGLp8 ( n &b* -- )
; SUBGLp16 ( n &w* -- )
; SUBGLp32 ( n &l* -- )

 macro SUBGL
 	pop	hl
	call	subgl
	pop	de
 endm
 	
 macro SUBGLp16
 	pop	hl
	call	subglp16
	pop	de
 endm
 	
 macro SUBGLp32
 	pop	hl
	call	subglp32
	pop	de
 endm
 	
subglp32::
	add	hl,hl
subglp16::
	add	hl,hl
subgl::	ld	a,(de)
	sub	l
	ld	(de++),a
	ld	a,(de)
	sbc	h
	ld	(de),a
	ret

; _________________________________
; divide unsigned variable
;
; DIVGLu ( n &w -- )

 macro DIVGLu
 	pop	hl
	call	divglu
	pop	de
 endm
 	
divglu::
	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,(hl++)
	ld	h,(hl)	
	ld	l,a		; hl=n1, de=n2

	call	divu		; de = hl / de
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; divide signed variable
;
; DIVGLs ( n &w -- )

 macro DIVGLs
 	pop	hl
	call	divgls
	pop	de
 endm
 	
divgls::
	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,(hl++)
	ld	h,(hl)	
	ld	l,a		; hl=n1, de=n2

	call	divs		; de = hl / de
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret


; _________________________________
; divide unsigned variable and store remainder
;
; REMGLu ( n &w -- )

 macro REMGLu
 	pop	hl
	call	remglu
	pop	de
 endm
 	
remglu::
	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,(hl++)
	ld	h,(hl)	
	ld	l,a		; hl=n1, de=n2

	call	divu		; hl = hl % de
	ex	hl,de
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; divide signed variable and store remainder
;
; REMGLs ( n &w -- )

 macro REMGLs
 	pop	hl
	call	remgls
	pop	de
 endm
 	
remgls::
	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,(hl++)
	ld	h,(hl)	
	ld	l,a		; hl=n1, de=n2

	call	rems		; hl = hl % de
	ex	hl,de
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; multiply value to variable
;
; MULGL ( n &w -- )

 macro MULGL
 	pop	hl
	call	mulgl
	pop	de
 endm
 
mulgl:: 	
	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,(hl++)
	ld	h,(hl)	
	ld	l,a		; hl=n1, de=n2

	call	mul_hl_de	; hl = hl * de
	ex	hl,de
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; shift left variable
;
; SLGL ( n &w -- )

 macro SLGL
 	pop	hl
	call	slgl
	pop	de
 endm

; shift left variable: (de) <<= l
slgl:: 	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,e		; a=shift
	ld	e,(hl++)
	ld	d,(hl)		; de=var, a=shift

	call	sl		; de = de << a
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; shift right signed variable
;
; SRGLs ( n &w -- )

 macro SRGLs
 	pop	hl
	call	srgls
	pop	de
 endm
 
srgls::	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,e		; a=shift
	ld	e,(hl++)
	ld	d,(hl)		; de=var, a=shift

	call	srs		; de = de >> a
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; _________________________________
; shift right unsigned variable
;
; SRGLu ( n4 &w -- )

 macro SRGLu
 	pop	hl
	call	srglu
	pop	de
 endm
 
srglu::	ex	hl,de		; hl -> var, de=value
	push	hl

	ld	a,e		; a=shift
	ld	e,(hl++)
	ld	d,(hl)		; de=var, a=shift

	call	sru		; de = de >> a
	
	pop	hl
	ld	(hl++),e
	ld	(hl),d
	ret

; ________________________________
; decrement word pointer variable
;
; MMSETp16 ( &w* -- )		

 macro MMSETp16
 	call	mmsetp16
	pop	de
 endm
 	
mmsetp16::
	ld	a,(de)
	sub	2
	ld	(de),a
	ret	nc
	ex	hl,de
	inc	hl
	dec	(hl)
	ret

; ________________________________
; increment word pointer variable
;
; PPSETp16 ( &w* -- )		

 macro PPSETp16
 	call	ppsetp16
	pop	de
 endm
 
ppsetp16::
	ld	a,(de)
	add	2
	ld	(de),a
	ret	nc
	ex	hl,de
	inc	hl
	inc	(hl)
	ret

; ________________________________
; increment variable
;
; PPSET   ( &w -- )
; PPSETp8 ( &b* -- )		; increment byte pointer

 macro PPSET
 	call	ppset
	pop	de
 endm	
 
ppset::	ex	de,hl		; hl->var	
	inc	(hl)
	ret	nz
	inc	hl
	inc	(hl)
	ret

; ________________________________
; decrement variable
;
; MMSET   ( &w -- )		; decrement word variable
; MMSETp8 ( &b* -- )		; decrement byte pointer

 macro MMSET
 	call	mmset
	pop	de
 endm
 
mmset::	ex	de,hl		; hl->var	
	dec	(hl)
	ld	a,(hl)
	inc	a
	ret	nz
	inc	hl
	dec	(hl)
	ret

; ________________________________
; get value from variable
; with post-increment
;
; PEEKPP   ( &w -- n )
; PEEKPPp8 ( &b* -- b* )	; peek byte pointer with post increment

 macro PEEKPP
 	call	peekpp
 endm
 	
peekpp::
	ex	hl,de		; hl->var
	ld	e,(hl++)
	ld	d,(hl)
	inc	de
	ld	(hl),d
	ld	(--hl),e
	dec	de
	ret

; ________________________________
; get value from variable
; with post-decrement
;
; PEEKMM   ( &w -- n )
; PEEKMMp8 ( &b* -- b* )	; peek byte pointer with post decrement

 macro PEEKMM
 	call	peekmm
 endm
 
peekmm:: 	
	ex	hl,de		; hl->var
	ld	e,(hl++)
	ld	d,(hl)
	dec	de
	ld	(hl),d
	ld	(--hl),e
	inc	de
	ret

; ________________________________
; get value from variable
; with pre-decrement
;
; MMPEEK   ( &w -- n )
; MMPEEKp8 ( &b* -- b* )	; peek bype pointer with pre decrement

 macro MMPEEK
 	call	mmpeek
 endm
 	
mmpeek::
	ex	de,hl		; hl->var
	ld	e,(hl++)
	ld	d,(hl)
	dec	de
	ld	(hl),d
	ld	(--hl),e
	ret

; ________________________________
; get value from variable
; with pre-increment
;
; PPPEEK   ( &w -- n )
; PPPEEKp8 ( &b* -- b* )	; peek byte pointer with pre-increment

 macro PPPEEK
 	call	pppeek
 endm
 	
pppeek::
	ex	de,hl		; hl->var
	ld	e,(hl++)
	ld	d,(hl)
	inc	de
	ld	(hl),d
	ld	(--hl),e
	ret

; ________________________________
; get value from variable
;
; PEEK ( &w -- n )

 macro PEEK
 	ex	hl,de		; hl -> var	
	ld	e,(hl++)
	ld	d,(hl)
 endm

; ________________________________
; store value in variable
;
; POKE ( n &w -- )

 macro POKE
	pop	hl
 	ex	hl,de		; hl -> var, de=value
	ld	(hl++),e
	ld	(hl),d
	pop	de
 endm	

; ________________________________
; store value in variable
;
; NPOKE ( &w n -- )

 macro NPOKE
 	pop	hl		; hl -> var
	ld	(hl++),e
	ld	(hl),d
	pop	de
 endm	


; ________________________________
; 16 bit Galois LFSR
;
; https://en.wikipedia.org/wiki/Linear_feedback_shift_register
;
; this example produces the bit stream 
; the used polynomial results in a period of 2^16-1
; IMPORTANT: seed must not be 0x0000
; 	     0x0000 is never returned!
;
; uint16 start_state = 0xACE1u;  /* Any nonzero start state will work. */
; uint16 lfsr = start_state;
; unsigned period = 0;
;
; do	
; {
;     unsigned lsb = lfsr & 1;   /* Get LSB (i.e., the output bit). */
;     lfsr >>= 1;                /* Shift register */
;     lfsr ^= (-lsb) & 0xB400u;  /* If the output bit is 1, apply toggle mask.
;                                 * The value has 1 at bits corresponding
;                                 * to taps, 0 elsewhere. */
;     ++period;
; } while (lfsr != start_state);
;
; wir schieben andersrum und das Polynom ist = $002D
; 0xB400 = 10110100 00000000 -> 00000000 00101101 = 0x002D
;
; RAND ( -- n )

 macro RAND
	push	de
	call	rand
 endm
	
rand::	ld	hl,(rndseed16)	; get LFSR
	
	ld	de,0x2D10	; d = polynomial.lo, e = counter for 16 bits
		
2$:	add	hl,hl
	jr	nc,1$
	ld	a,l
	xor	d
	ld	l,a
1$:	dec	e
	jr	nz,2$
		
	ld	(rndseed16),hl	; save LFSR
	
	dec	hl		; range [0xffff..0x0001] -> [0xfffe..0x0000]
	ex	hl,de
	ret


; ________________________________
; swap two word variables
;
; SWAPVARS ( &var &var -- )

 macro SWAPVARS
	pop	hl
	call	swapvars
	pop	de
 endm

swapvars32::
	ld	a,(de)
	ldi
	dec	hl		; undo
	ld	(hl++),a

	ld	a,(de)
	ldi
	dec	hl		; undo
	ld	(hl++),a

	inc	bc		; undo
	inc	bc		; undo

swapvars::	
	ld	a,(de)
	ldi
	dec	hl		; undo
	ld	(hl++),a

	ld	a,(de)
	ldi
	dec	hl		; undo
	ld	(hl),a		

	inc	bc		; undo
	inc	bc		; undo
	ret





#endlocal

